# -*- coding: utf-8 -*-
"""
Created on Wed Aug 22 15:06:52 2018
@author: Matthew Shahnazari

Updated by The Lantau Group (Tom Wilson, Paul Buckland)
"""

# Import Python libraries
# Also need 'openpyxl' installed
import pandas as pd
from pandas import ExcelWriter
import time
import numpy as np
import os
import re

# #### SET-UP #####

# Retrieve all 'Load_' Excel files
regex = re.compile('^Load_.*xls.$')
loadFiles = []
files = [f for f in os.listdir('.') if os.path.isfile(f)]
for f in files:
    if regex.match(f):
        loadFiles.append(f)
if len(loadFiles) == 0:
    print('*** ERROR: No load files found. Aborting')
    quit()

# Alert the user
print('\nThe following load files will be processed:')
for st in loadFiles:
    print(' - ' + st)

# Choices for demand data
Demand_Data_Options = ["Baseline Demand Profile rounded",
                       "Net Demand Profile rounded"]

# Output worksheet names to use
Demand_Data_TabNames = ["Baseline Demand Profile rounded",
                        "Net Demand Profile rounded"]

# Set up LOLE dataframe
LOLE_results = pd.DataFrame(np.nan, index=['LOLP Baseline Demand Profile',
                                           'LOLP Net Demand Profile'], columns=['LOLE (TIs)',
                                                            'LOLE (Hrs)',
                                                            'ELCC'])

# Alert the user
print('\nLOLP will be calculated for the following:')
for st in Demand_Data_Options:
    print(' - ' + st)

# Start timing
startTime = time.time()

# Read Excel data into data frames
COPT_Data = pd.read_excel("Working_COPT.xlsx", header=0, index_col=None, engine='openpyxl')


# Define LOLP Maker
def LOLP_Maker(data):

    # Assumes that 'outageCapacity' in teh 'Working_COPT.xlsx' file is 0 to CC

    # Index to find cumulative probability
    reserveT = (CC - loadTbl[data]).map(lambda x: int(x))
    reserveT_0 = reserveT.map(lambda x: max(0, x)).to_list()

    # Set LOLP
    forLOLP = COPT_Data['cumulativeProbability'].iloc[reserveT_0]
    LOLP_output = pd.DataFrame({'Load': loadTbl[data].tolist(), 'LOLP': forLOLP.to_list()})

    # If reserveT -ve
    LOLP_output.loc[(reserveT < 0), 'Load'] = 0
    LOLP_output.loc[(reserveT < 0), 'LOLP'] = 1

    return LOLP_output


# Start main loop
for loadFile in loadFiles:

    # Alert the user
    print('\nProcessing load file: ' + loadFile)
    print('=' * (len(loadFile)+22))

    # Read in load file
    loadTbl = pd.read_excel(loadFile, header=0, engine='openpyxl')

    # Open the output Excel file (it's deleted if it already exists)
    writer = ExcelWriter('Output_LOLP_' + loadFile + '.xlsx', engine='openpyxl')

    # Total installed capacity of conventional generators, read from COPT
    CC = COPT_Data.iloc[0]['availableCapacity']

    LOLP_actual_col = LOLP_Maker('Baseline Demand Profile rounded')
    LOLE_actual = LOLP_actual_col['LOLP'].sum()

    for UserChoice in range(1, 3):

        # Demand profile defined as per the user's choice
        Demand_Data_System = 'Baseline Demand Profile rounded'
        Net_Demand_data = Demand_Data_Options[UserChoice - 1]
        Net_Demand_TabName = Demand_Data_TabNames[UserChoice - 1]

        # Alert the user
        print('Processing: ' + Net_Demand_data)

        # Total installed capacity of conventional generators, read from COPT
        CC = COPT_Data.iloc[0]['availableCapacity']

        # Find number of rows of input file to define output file
        LOLP_rowCount = len(loadTbl)

        # #### OFFSET FINDER LOOP #####

        # Upper limit depends on UserChoice, to minimise required iterations
        # Different to base case because we need to first get up to target LOLE
        Upper = 10000
        MaxN = 100

        # Define other required variables
        Lower = 0
        c = 0
        TOL = 0.001 * LOLE_actual
        N = 0

        # While loop uses bisection method to find target offset MW
        while N < MaxN:

            loadTbl[Net_Demand_data] = loadTbl[Net_Demand_data] - c + Lower
            LOLPa = LOLP_Maker(Net_Demand_data)
            f_a = LOLPa['LOLP'].sum()

            loadTbl[Net_Demand_data] = loadTbl[Net_Demand_data] - Lower + Upper
            LOLPb = LOLP_Maker(Net_Demand_data)
            f_b = LOLPb['LOLP'].sum()

            c = int(round(((Upper + Lower) / 2)))
            loadTbl[Net_Demand_data] = loadTbl[Net_Demand_data] - Upper + c
            LOLPc = LOLP_Maker(Net_Demand_data)
            f_c = LOLPc['LOLP'].sum()

            # Heuristic used to decide which to upper/lower bounds to pass back through the loop
            # If target offset found, then define results and exit loop
            # If error, exit loop to prevent infinite or pointless looping
            if (abs(f_c - LOLE_actual) < TOL) or (N == (MaxN-1)):
                Result_Table = LOLPc
                Result_Offset = int(c)
                break
            elif f_c > LOLE_actual:
                Lower = Lower
                Upper = c
            elif f_c < LOLE_actual:
                Upper = Upper
                Lower = c
            else:
                break

            N = N + 1

        # #### OUTPUT OF RESULTS #####

        # Define column of datetimes, as per the loadTbl (loaded from the input sheet)
        # Add 'Interval' column in output sheet with these datetimes
        datetimeTbl = loadTbl['Trading_Interval']
        Result_Table['Interval'] = datetimeTbl

        # Write output file to excel
        # writer = ExcelWriter("Output_LOLP.xlsx", engine='openpyxl')
        Result_Table.to_excel(writer, Net_Demand_TabName, index=False)
        writer.save()

        # LOLE results
        LOLE_System_24hr_result = Result_Table['LOLP'].sum()
        LOLE_hr_System_24hr_result = LOLE_System_24hr_result/2
        LOLE_results.iloc[UserChoice - 1, 0] = LOLE_System_24hr_result
        LOLE_results.iloc[UserChoice - 1, 1] = LOLE_hr_System_24hr_result
        LOLE_results.iloc[UserChoice - 1, 2] = Result_Offset

    # Write LOLE results (once all demand types have been calculated
    LOLE_results.to_excel(writer, 'Results', index=True)
    writer.save()

# End timer
print('\nTime taken: %s seconds' % (time.time()-startTime))
